
#define FIXED_TO_FLOAT(x) ((float)(x >> 16));
#define FLOAT_TO_FIXED(f) (long)(f * 65536);
#define INT_TO_FIXED(x) (long)(x << 16)
#define FIXED_TO_INT(x) (int) ( (long)(x >> 16) & 0x0000ffff)

/*  long long -> turns into 64-bit
    Any 32-bit CPU or even MMX will go ape shit over this
    and lose any speed gain from fixed*/
//#define FIXEDMUL(x, y) ((long long)(x)*(long long)(y)) / 65536

/*  Not super accurate but good enough*/
#define FIXEDMUL(x, y) ((x>>8)*(y>>8))

//#define FIXEDDIV(x, y) (((long long)(x)<<16)/y)
#define FIXEDDIV(x, y) ((x<<5)/y)<<11 

struct Vec {
   long x, y, z;
}; 
#define FLOAT_TO_FIXED(f) (long)(f * 65536);
#define FIXEDMUL(x, y) ((x>>8)*(y>>8))
inline long FIXCEIL( long a) {
    long decimal, integer;
    decimal = a & 0x0000ffff;
    if(decimal != 0) {
        integer = a & 0xffff0000;
        integer += 0xffff0000;
        return integer;
    }
    return a;
}
extern long *SIN1024;
extern long *COS1024;
inline struct Vec VecCreateF( float x_, float y_, float z_) {
    struct Vec r_;
    r_.x = FLOAT_TO_FIXED(x_);
    r_.y = FLOAT_TO_FIXED(y_);
    r_.z = FLOAT_TO_FIXED(z_);
    return r_;
};
struct Vec VecCreateL( long x_, long y_, long z_);
inline void add( struct Vec *A, struct Vec *B) {
    (*A).x+=(*B).x;
    (*A).y+=(*B).y;
    (*A).z+=(*B).z;
}
struct Vec sub( struct Vec A, struct Vec B);
struct Vec mul( struct Vec A, struct Vec B);
struct Vec norm( struct Vec A);
inline float dot( struct Vec A, struct Vec B) {
    return A.x*B.x+A.y*B.y+A.z*B.z;
}
struct Vec cross( struct Vec A, struct Vec B);
inline void rot2D( struct Vec *A, long rad ) {
    long X, Y;
    long sina = SIN1024[rad&1023];
    long cosa = COS1024[rad&1023];
    X = FIXEDMUL((*A).x , cosa ) - FIXEDMUL((*A).y , sina);
    Y = FIXEDMUL((*A).x , sina ) + FIXEDMUL((*A).y , cosa);
    (*A).x=X;
    (*A).y=Y;
}


#define zbufferlen 320*200*2
#define vertexarraylen 30000
#define colorarraylen 60000
extern dword *zbuffer;
extern long *vertexarray;
extern byte *colorarray;
dword zbufferpointer = 0;
dword vertexarraypointer = 0;
dword colorarraypointer = 0;
void initVector() {
    zbuffer = malloc(4*zbufferlen);
    vertexarray = malloc(sizeof(long)*vertexarraylen);
    colorarray = malloc(sizeof(byte)*colorarraylen);
}

inline void drawHorizontalPoly(word x0, word x1, word y, byte color, long startz, long stepz) {
    word pointer8, length8;
    word i;
    if(y < 0) return;
    if(y > 200) return;
    if(horzlbufindx>horzlbufferlen) return;
    if( x0 > x1 ) { //if bigger xor-swap
        x0 ^= x1;
        x1 ^= x0;
        x0 ^= x1;
    }
    pointer8 = y * 320 + x0;
    length8 = x1 - x0;
    x0 = pointer8 % 320;
    if(length8>0)
    for ( ;length8 > 0;length8--) {
        if(x0 > 0 && x0 < 318) {
            startz += stepz;
            if( zbuffer[ pointer8 ] > startz) {
                videobuffer[ pointer8 ] = color;
                zbuffer[ pointer8 ] = startz;
            }
        }
        pointer8++;
        x0++;
    }
}
inline static void drawPolygonZbuffered(long x1, long y1, long z1, long x2, long y2, long z2, long x3, long y3, long z3) {

    long invslope1, invslope2, curx1, curx2;
    long scanlineY;

    long centerX = 160 * 65536;
    long centerY = 160 * 65536;

    long curZ_A, curZ_B, changeZ_A, changeZ_B;
    long rasterZstep;

    curZ_A = z1;
    curZ_B = z1;
    changeZ_A = FIXEDDIV( z2 - z1, abs( y2 - y1 ) );
    changeZ_B = FIXEDDIV( z3 - z1, abs( y3 - y1 ) );
    x1 = FIXEDDIV(x1, z1);
    y1 = FIXEDDIV(y1, z1);
    x2 = FIXEDDIV(x2, z2);
    y2 = FIXEDDIV(y2, z2);
    x3 = FIXEDDIV(x3, z3);
    y3 = FIXEDDIV(y3, z3);
    x1 += centerX;
    x2 += centerX;
    x3 += centerX;
    y1 += centerY;
    y2 += centerY;
    y3 += centerY;
    x1 = FIXCEIL(x1);
    y1 = FIXCEIL(y1);
    x2 = FIXCEIL(x2);
    y2 = FIXCEIL(y2);
    x3 = FIXCEIL(x3);
    y3 = FIXCEIL(y3);
    
    if(y1>y2) {
        swap( &x1, &x2);
        swap( &y1, &y2);
    }
    if(y1>y3) {
        swap( &x1, &x3);
        swap( &y1, &y3);
    }
    if(y2>y3) {
        swap( &x2, &x3);
        swap( &y2, &y3);
    }

    /*  Y coordinate has to be relocated to make fixed point division work better. 
        And afterwards relocated back to its orignal position. This should be fixed!
        Causes loss of decent bunch cycles for every polygon, not terribly bad but would improve results. */
    y1 += 128 * 65536;
    y2 += 128 * 65536;
    y3 += 128 * 65536;


    if(y1<0) y1 = 256;
    if(y2<0) y2 = 256;
    if(y3<0) y3 = 256;

    curx1 = x1;
    curx2 = x1;
    scanlineY = y1;

    if(y1 != y2) {
        invslope1 = FIXEDDIV( (x2 - x1) , (y2 - y1) );
        invslope2 = FIXEDDIV( (x3 - x1) , (y3 - y1) );

        for ( scanlineY = FIXED_TO_INT(y1); scanlineY < FIXED_TO_INT(y2); scanlineY++) {
            int ix0 = FIXED_TO_INT(curx1);
            int ix1 = FIXED_TO_INT(curx2);
            long Z_Ai = FIXEDMUL ( changeZ_A , invslope1 );
            long Z_Bi = FIXEDMUL ( changeZ_B , invslope2 );
            rasterZstep = FIXEDMUL ( curZ_B - curZ_A , invslope1 );
            drawHorizontalPoly( ix0, ix1, scanlineY-128/*relocation back to real position*/, 33, curZ_A, 0);
            curx1 += invslope1;
            curx2 += invslope2;
            if(scanlineY>200+128)
                break;
        }
    }

    if(y1 == y2) curx2 = x2;
    curx1 = curx1;
    curx2 = curx2;

    changeZ_A = FIXEDDIV( z2 - curZ_A, abs(y2-scanlineY));
    changeZ_B = FIXEDDIV( z3 - curZ_B, abs(y3-scanlineY));

    if(y2 != y3) {
        long fscnly=INT_TO_FIXED ( scanlineY );
        invslope1 = FIXEDDIV( (x3 - curx1) , (y3 - fscnly ) );
        invslope2 = FIXEDDIV( (x3 - curx2) , (y3 - fscnly ) );

        for ( scanlineY = scanlineY; scanlineY < FIXED_TO_INT(y3); scanlineY++) {
            int ix0 = FIXED_TO_INT(curx1);
            int ix1 = FIXED_TO_INT(curx2);
            curZ_A += changeZ_A;
            curZ_B += changeZ_B;
            rasterZstep = FIXEDDIV ( curZ_B - curZ_A , (curx2 - curx1) );
            drawHorizontalPoly( ix0, ix1, scanlineY-128, 99, curZ_A, 0);
            curx1 += invslope1;
            curx2 += invslope2;
            if(scanlineY>200+128)
                break;
        }
    }

}

inline static void iglBegin( ) {
    dword i;
    for(i=0;i<320*200*2; i++) {
        zbuffer[i] = 255;
    }
    vertexarraypointer = 0;
    colorarraypointer = 0;
}
inline static void iglEnd( ) {
    dword i, a;
    a=0;
    drawPolygonZbuffered(-65536*32, -65536*32, 65536*1 , 65536*32, -65536*32, 65536*1, 0, -65536*32, 65536*1);
    
}
inline static void iglColor( byte c ) {
    colorarray[colorarraypointer++] = c;
}
inline static void iglVertex3( struct Vec v) {
    vertexarray[vertexarraypointer++] = v.x;
    vertexarray[vertexarraypointer++] = v.y;
    vertexarray[vertexarraypointer++] = v.z;
}
inline static void iglVertex3l( long x, long y, long z) {
    vertexarray[vertexarraypointer++] = x;
    vertexarray[vertexarraypointer++] = y;
    vertexarray[vertexarraypointer++] = z;
}
inline static void iglVertex3f( float x, float y, float z) {
    long lx = FLOAT_TO_FIXED(x);
    long ly = FLOAT_TO_FIXED(y);
    long lz = FLOAT_TO_FIXED(z);
    vertexarray[vertexarraypointer++] = lx;
    vertexarray[vertexarraypointer++] = ly;
    vertexarray[vertexarraypointer++] = lz;
}